@@ -779,7 +779,7 @@ static int connection_handle_read_state(connection * const con)  {
     connection_set_state(r, CON_STATE_REQUEST_END);
 
     if (light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE)
-        && !con->is_ssl_sock && r->conf.h2proto && 0 == r->http_status
+        && 0 == r->http_status
         && h2_check_con_upgrade_h2c(r)) {
         /*(Upgrade: h2c over cleartext does not have SNI; no COMP_HTTP_HOST)*/
         r->conditional_is_valid = (1 << COMP_SERVER_SOCKET)
@@ -1275,17 +1275,21 @@ connection_state_machine_h2 (request_st * const h2r, connection * const con)
 
                 uint32_t dlen = max_bytes > 32768 ? 32768 : (uint32_t)max_bytes;
                 dlen = h2_send_cqdata(r, con, &r->write_queue, dlen);
-                max_bytes -= (off_t)dlen;
+                if (dlen) { /*(do not resched (spin) if swin empty window)*/
+                    max_bytes -= (off_t)dlen;
+                    if (!chunkqueue_is_empty(&r->write_queue))
+                        resched |= 1;
+                }
+            }
 
+            {
                 if (chunkqueue_is_empty(&r->write_queue)) {
-                    if (r->resp_body_finished) {
+                    if (r->resp_body_finished && r->state == CON_STATE_WRITE) {
                         connection_set_state(r, CON_STATE_RESPONSE_END);
                         if (__builtin_expect( (r->conf.log_state_handling), 0))
                             connection_state_machine_loop(r, con);
                     }
                 }
-                else if (dlen) /*(do not spin if swin empty window)*/
-                    resched |= 1; /*(!chunkqueue_is_empty(&r->write_queue))*/
             }
 
           #if 0
@@ -1299,7 +1303,7 @@ connection_state_machine_h2 (request_st * const h2r, connection * const con)
                 /*(trigger reschedule of con if frames pending)*/
                 if (h2c->rused == sizeof(h2c->r)/sizeof(*h2c->r)
                     && !chunkqueue_is_empty(con->read_queue))
-                    resched |= 1;
+                    resched |= 2;
                 h2_send_end_stream(r, con);
                 const int alive = r->keep_alive;
                 h2_retire_stream(r, con);/*r invalidated;removed from h2c->r[]*/
@@ -1313,7 +1317,7 @@ connection_state_machine_h2 (request_st * const h2r, connection * const con)
             }
         }
 
-        if (0 == max_bytes) resched |= 2;
+        if (0 == max_bytes) resched |= 1;
     }
 
     if (h2c->sent_goaway > 0 && h2c->rused) {
@@ -1339,12 +1343,15 @@ connection_state_machine_h2 (request_st * const h2r, connection * const con)
         if (chunkqueue_is_empty(con->write_queue)
             && 0 == h2c->rused && h2c->sent_goaway)
             connection_set_state(h2r, CON_STATE_RESPONSE_END);
-        else if (resched & 2)
-            resched = (resched & 1) | (chunkqueue_is_empty(con->write_queue));
     }
 
     if (h2r->state == CON_STATE_WRITE) {
-        if (resched && !con->traffic_limit_reached)
+        /* (resched & 1) more data is available to write, if still able to write
+         * (resched & 2) resched to read deferred frames from con->read_queue */
+        /*(con->is_writable set to 0 if !chunkqueue_is_empty(con->write_queue)
+         * after trying to write in connection_handle_write() above)*/
+        if (((resched & 1) && con->is_writable>0 && !con->traffic_limit_reached)
+            || (resched & 2))
             joblist_append(con);
 
         if (h2_want_read(con))
