@@ -3,9 +3,10 @@
 #include "base.h"
 #include "stat_cache.h"
 #include "http_kv.h"
+#include "fdlog.h"
 #include "log.h"
-#include "connections.h"
 #include "response.h"
+#include "http_cgi.h"
 #include "http_chunk.h"
 #include "http_header.h"
 
@@ -26,28 +27,12 @@
 #include <fcntl.h>
 #include <signal.h>
 
-static int pipe_cloexec(int pipefd[2]) {
-  #ifdef HAVE_PIPE2
-    if (0 == pipe2(pipefd, O_CLOEXEC)) return 0;
-  #endif
-    return 0 == pipe(pipefd)
-       #ifdef FD_CLOEXEC
-        && 0 == fcntl(pipefd[0], F_SETFD, FD_CLOEXEC)
-        && 0 == fcntl(pipefd[1], F_SETFD, FD_CLOEXEC)
-       #endif
-      ?  0
-      : -1;
-}
-
 typedef struct {
-	char *ptr;
-	size_t used;
-	size_t size;
-	size_t *offsets;
+	uintptr_t *offsets;
 	size_t osize;
 	size_t oused;
-	char **eptr;
-	size_t esize;
+	buffer *b;
+	buffer *boffsets;
 	buffer *ld_preload;
 	buffer *ld_library_path;
       #ifdef __CYGWIN__
@@ -56,13 +41,14 @@ typedef struct {
 } env_accum;
 
 typedef struct {
-	struct { pid_t pid; void *ctx; } *ptr;
-	size_t used;
-	size_t size;
-} buffer_pid_t;
+	unix_time64_t read_timeout;
+	unix_time64_t write_timeout;
+	int signal_fin;
+} cgi_limits;
 
 typedef struct {
 	const array *cgi;
+	const cgi_limits *limits;
 	unsigned short execute_x_only;
 	unsigned short local_redir;
 	unsigned short xsendfile_allow;
@@ -70,31 +56,47 @@ typedef struct {
 	const array *xsendfile_docroot;
 } plugin_config;
 
+struct cgi_pid_t;
+
 typedef struct {
 	PLUGIN_DATA;
 	plugin_config defaults;
 	plugin_config conf;
-	buffer_pid_t cgi_pid;
+	int tempfile_accum;
+	struct cgi_pid_t *cgi_pid;
 	env_accum env;
 } plugin_data;
 
 typedef struct {
-	pid_t pid;
+	struct cgi_pid_t *cgi_pid;
 	int fd;
 	int fdtocgi;
+	int rd_revents;
+	int wr_revents;
 	fdnode *fdn;
 	fdnode *fdntocgi;
 
 	request_st *r;
+	connection *con;          /* dumb pointer */
 	struct fdevents *ev;      /* dumb pointer */
 	plugin_data *plugin_data; /* dumb pointer */
 
 	buffer *response;
+	unix_time64_t read_ts;
+	unix_time64_t write_ts;
 	buffer *cgi_handler;      /* dumb pointer */
 	http_response_opts opts;
 	plugin_config conf;
 } handler_ctx;
 
+typedef struct cgi_pid_t {
+	pid_t pid;
+	int signal_sent;
+	handler_ctx *hctx;
+	struct cgi_pid_t *next;
+	struct cgi_pid_t *prev;
+} cgi_pid_t;
+
 static handler_ctx * cgi_handler_ctx_init(void) {
 	handler_ctx *hctx = calloc(1, sizeof(*hctx));
 
@@ -137,16 +139,32 @@ INIT_FUNC(mod_cgi_init) {
 
 FREE_FUNC(mod_cgi_free) {
 	plugin_data *p = p_d;
-	buffer_pid_t *bp = &(p->cgi_pid);
-	if (bp->ptr) free(bp->ptr);
-	free(p->env.ptr);
-	free(p->env.offsets);
-	free(p->env.eptr);
 	buffer_free(p->env.ld_preload);
 	buffer_free(p->env.ld_library_path);
       #ifdef __CYGWIN__
 	buffer_free(p->env.systemroot);
       #endif
+
+    for (cgi_pid_t *cgi_pid = p->cgi_pid, *next; cgi_pid; cgi_pid = next) {
+        next = cgi_pid->next;
+        free(cgi_pid);
+    }
+
+    if (NULL == p->cvlist) return;
+    /* (init i to 0 if global context; to 1 to skip empty global context) */
+    for (int i = !p->cvlist[0].v.u2[1], used = p->nconfig; i < used; ++i) {
+        config_plugin_value_t *cpv = p->cvlist + p->cvlist[i].v.u2[0];
+        for (; -1 != cpv->k_id; ++cpv) {
+            if (cpv->vtype != T_CONFIG_LOCAL || NULL == cpv->v.v) continue;
+            switch (cpv->k_id) {
+              case 6: /* cgi.limits */
+                free(cpv->v.v);
+                break;
+              default:
+                break;
+            }
+        }
+    }
 }
 
 static void mod_cgi_merge_config_cpv(plugin_config * const pconf, const config_plugin_value_t * const cpv) {
@@ -169,6 +187,10 @@ static void mod_cgi_merge_config_cpv(plugin_config * const pconf, const config_p
       case 5: /* cgi.upgrade */
         pconf->upgrade = (unsigned short)cpv->v.u;
         break;
+      case 6: /* cgi.limits */
+        if (cpv->vtype != T_CONFIG_LOCAL) break;
+        pconf->limits = cpv->v.v;
+        break;
       default:/* should not happen */
         return;
     }
@@ -189,6 +211,96 @@ static void mod_cgi_patch_config(request_st * const r, plugin_data * const p) {
     }
 }
 
+__attribute_cold__
+__attribute_pure__
+static int mod_cgi_str_to_signal (const char *s, int default_sig) {
+    static const struct { const char *name; int sig; } sigs[] = {
+      { "HUP",  SIGHUP  }
+     ,{ "INT",  SIGINT  }
+     ,{ "QUIT", SIGQUIT }
+     ,{ "ILL",  SIGILL  }
+     ,{ "TRAP", SIGTRAP }
+     ,{ "ABRT", SIGABRT }
+     #ifdef SIGBUS
+     ,{ "BUS",  SIGBUS  }
+     #endif
+     ,{ "FPE",  SIGFPE  }
+     ,{ "KILL", SIGKILL }
+     #ifdef SIGUSR1
+     ,{ "USR1", SIGUSR1 }
+     #endif
+     ,{ "SEGV", SIGSEGV }
+     #ifdef SIGUSR2
+     ,{ "USR2", SIGUSR2 }
+     #endif
+     ,{ "PIPE", SIGPIPE }
+     ,{ "ALRM", SIGALRM }
+     ,{ "TERM", SIGTERM }
+     #ifdef SIGCHLD
+     ,{ "CHLD", SIGCHLD }
+     #endif
+     #ifdef SIGCONT
+     ,{ "CONT", SIGCONT }
+     #endif
+     #ifdef SIGURG
+     ,{ "URG",  SIGURG  }
+     #endif
+     #ifdef SIGXCPU
+     ,{ "XCPU", SIGXCPU }
+     #endif
+     #ifdef SIGXFSZ
+     ,{ "XFSZ", SIGXFSZ }
+     #endif
+     #ifdef SIGWINCH
+     ,{ "WINCH",SIGWINCH}
+     #endif
+     #ifdef SIGPOLL
+     ,{ "POLL", SIGPOLL }
+     #endif
+     #ifdef SIGIO
+     ,{ "IO",   SIGIO   }
+     #endif
+    };
+
+    if (s[0] == 'S' && s[1] == 'I' && s[2] == 'G') s += 3; /*("SIG" prefix)*/
+    for (uint32_t i = 0; i < sizeof(sigs)/sizeof(*sigs); ++i) {
+        if (0 == strcmp(s, sigs[i].name)) return sigs[i].sig;
+    }
+    return default_sig;
+}
+
+static cgi_limits * mod_cgi_parse_limits(const array * const a, log_error_st * const errh) {
+    cgi_limits * const limits = calloc(1, sizeof(cgi_limits));
+    force_assert(limits);
+    for (uint32_t i = 0; i < a->used; ++i) {
+        const data_unset * const du = a->data[i];
+        int32_t v = config_plugin_value_to_int32(du, -1);
+        if (buffer_eq_icase_slen(&du->key, CONST_STR_LEN("read-timeout"))) {
+            limits->read_timeout = (unix_time64_t)v;
+            continue;
+        }
+        if (buffer_eq_icase_slen(&du->key, CONST_STR_LEN("write-timeout"))) {
+            limits->write_timeout = (unix_time64_t)v;
+            continue;
+        }
+        if (buffer_eq_icase_slen(&du->key, CONST_STR_LEN("tcp-fin-propagate"))) {
+            if (-1 == v) {
+                v = SIGTERM;
+                if (du->type == TYPE_STRING) {
+                    buffer * const vstr = &((data_string *)du)->value;
+                    buffer_to_upper(vstr);
+                    v = mod_cgi_str_to_signal(vstr->ptr, SIGTERM);
+                }
+            }
+            limits->signal_fin = v;
+            continue;
+        }
+        log_error(errh, __FILE__, __LINE__,
+                  "unrecognized cgi.limits param: %s", du->key.ptr);
+    }
+    return limits;
+}
+
 SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
     static const config_plugin_keys_t cpk[] = {
       { CONST_STR_LEN("cgi.assign"),
@@ -209,6 +321,9 @@ SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
      ,{ CONST_STR_LEN("cgi.upgrade"),
         T_CONFIG_BOOL,
         T_CONFIG_SCOPE_CONNECTION }
+     ,{ CONST_STR_LEN("cgi.limits"),
+        T_CONFIG_ARRAY_KVANY,
+        T_CONFIG_SCOPE_CONNECTION }
      ,{ NULL, 0,
         T_CONFIG_UNSET,
         T_CONFIG_SCOPE_UNSET }
@@ -221,7 +336,7 @@ SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
     /* process and validate config directives
      * (init i to 0 if global context; to 1 to skip empty global context) */
     for (int i = !p->cvlist[0].v.u2[1]; i < p->nconfig; ++i) {
-        const config_plugin_value_t *cpv = p->cvlist + p->cvlist[i].v.u2[0];
+        config_plugin_value_t *cpv = p->cvlist + p->cvlist[i].v.u2[0];
         for (; -1 != cpv->k_id; ++cpv) {
             switch (cpv->k_id) {
               case 0: /* cgi.assign */
@@ -237,13 +352,18 @@ SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
                           cpk[cpv->k_id].k, ds->value.ptr);
                         return HANDLER_ERROR;
                     }
-                    buffer_path_simplify(&ds->value, &ds->value);
+                    buffer_path_simplify(&ds->value);
                     buffer_append_slash(&ds->value);
                 }
                 break;
               case 4: /* cgi.local-redir */
               case 5: /* cgi.upgrade */
                 break;
+              case 6: /* cgi.limits */
+                cpv->v.v = mod_cgi_parse_limits(cpv->v.a, srv->errh);
+                if (NULL == cpv->v.v) return HANDLER_ERROR;
+                cpv->vtype = T_CONFIG_LOCAL;
+                break;
               default:/* should not happen */
                 break;
             }
@@ -257,42 +377,39 @@ SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
             mod_cgi_merge_config(&p->defaults, cpv);
     }
 
+    p->tempfile_accum = config_feature_bool(srv, "cgi.tempfile-accum", 1);
+
     return HANDLER_GO_ON;
 }
 
 
-static void cgi_pid_add(plugin_data *p, pid_t pid, void *ctx) {
-    buffer_pid_t *bp = &(p->cgi_pid);
-
-    if (bp->used == bp->size) {
-        bp->size += 16;
-        bp->ptr = realloc(bp->ptr, sizeof(*bp->ptr) * bp->size);
-        force_assert(bp->ptr);
-    }
-
-    bp->ptr[bp->used].pid = pid;
-    bp->ptr[bp->used].ctx = ctx;
-    ++bp->used;
+static cgi_pid_t * cgi_pid_add(plugin_data *p, pid_t pid, handler_ctx *hctx) {
+    cgi_pid_t *cgi_pid = malloc(sizeof(cgi_pid_t));
+    force_assert(cgi_pid);
+    cgi_pid->pid = pid;
+    cgi_pid->signal_sent = 0;
+    cgi_pid->hctx = hctx;
+    cgi_pid->prev = NULL;
+    cgi_pid->next = p->cgi_pid;
+    p->cgi_pid = cgi_pid;
+    return cgi_pid;
 }
 
-static void cgi_pid_kill(plugin_data *p, pid_t pid) {
-    buffer_pid_t *bp = &(p->cgi_pid);
-    for (size_t i = 0; i < bp->used; ++i) {
-        if (bp->ptr[i].pid == pid) {
-            bp->ptr[i].ctx = NULL;
-            kill(pid, SIGTERM);
-            return;
-        }
-    }
+static void cgi_pid_kill(cgi_pid_t *cgi_pid, int sig) {
+    cgi_pid->signal_sent = sig; /*(save last signal sent)*/
+    kill(cgi_pid->pid, sig);
 }
 
-static void cgi_pid_del(plugin_data *p, size_t i) {
-    buffer_pid_t *bp = &(p->cgi_pid);
+static void cgi_pid_del(plugin_data *p, cgi_pid_t *cgi_pid) {
+    if (cgi_pid->prev)
+        cgi_pid->prev->next = cgi_pid->next;
+    else
+        p->cgi_pid = cgi_pid->next;
 
-    if (i != bp->used - 1)
-        bp->ptr[i] = bp->ptr[bp->used - 1];
+    if (cgi_pid->next)
+        cgi_pid->next->prev = cgi_pid->prev;
 
-    --bp->used;
+    free(cgi_pid);
 }
 
 
@@ -327,15 +444,13 @@ static void cgi_connection_close(handler_ctx *hctx) {
 	}
 
 	plugin_data * const p = hctx->plugin_data;
-
-	if (hctx->pid > 0) {
-		cgi_pid_kill(p, hctx->pid);
-	}
-
 	request_st * const r = hctx->r;
-
 	r->plugin_ctx[p->id] = NULL;
 
+	if (hctx->cgi_pid) {
+		cgi_pid_kill(hctx->cgi_pid, SIGTERM);
+		hctx->cgi_pid->hctx = NULL;
+	}
 	cgi_handler_ctx_free(hctx);
 
 	/* finish response (if not already r->resp_body_started, r->resp_body_finished) */
@@ -345,11 +460,13 @@ static void cgi_connection_close(handler_ctx *hctx) {
 }
 
 static handler_t cgi_connection_close_callback(request_st * const r, void *p_d) {
-	plugin_data *p = p_d;
-	handler_ctx *hctx = r->plugin_ctx[p->id];
-	if (hctx) cgi_connection_close(hctx);
-
-	return HANDLER_GO_ON;
+    handler_ctx *hctx = r->plugin_ctx[((plugin_data *)p_d)->id];
+    if (hctx) {
+        chunkqueue_set_tempdirs(&r->reqbody_queue, /* reset sz */
+                                r->reqbody_queue.tempdirs, 0);
+        cgi_connection_close(hctx);
+    }
+    return HANDLER_GO_ON;
 }
 
 
@@ -358,11 +475,13 @@ static int cgi_write_request(handler_ctx *hctx, int fd);
 
 static handler_t cgi_handle_fdevent_send (void *ctx, int revents) {
 	handler_ctx *hctx = ctx;
-	request_st * const r = hctx->r;
+	hctx->wr_revents |= revents;
+	joblist_append(hctx->con);
+	return HANDLER_FINISHED;
+}
 
-	/*(joblist only actually necessary here in mod_cgi fdevent send if returning HANDLER_ERROR)*/
-	joblist_append(r->con);
 
+static handler_t cgi_process_wr_revents (handler_ctx * const hctx, request_st * const r, int revents) {
 	if (revents & FDEVENT_OUT) {
 		if (0 != cgi_write_request(hctx, hctx->fdtocgi)) {
 			cgi_connection_close(hctx);
@@ -391,7 +510,7 @@ static handler_t cgi_handle_fdevent_send (void *ctx, int revents) {
 		return HANDLER_ERROR;
 	}
 
-	return HANDLER_FINISHED;
+	return HANDLER_GO_ON;
 }
 
 
@@ -429,10 +548,25 @@ static handler_t cgi_response_headers(request_st * const r, struct http_response
 }
 
 
+__attribute_cold__
+__attribute_noinline__
+static handler_t cgi_local_redir(request_st * const r, handler_ctx * const hctx) {
+    buffer_clear(hctx->response);
+    chunk_buffer_yield(hctx->response);
+    http_response_reset(r); /*(includes r->http_status = 0)*/
+    plugins_call_handle_request_reset(r);
+    /*cgi_connection_close(hctx);*//*(already cleaned up and hctx is now invalid)*/
+    return HANDLER_COMEBACK;
+}
+
+
 static int cgi_recv_response(request_st * const r, handler_ctx * const hctx) {
+		const off_t bytes_in = r->write_queue.bytes_in;
 		switch (http_response_read(r, &hctx->opts,
 					   hctx->response, hctx->fdn)) {
 		default:
+			if (r->write_queue.bytes_in > bytes_in)
+				hctx->read_ts = log_monotonic_secs;
 			return HANDLER_GO_ON;
 		case HANDLER_ERROR:
 			http_response_backend_error(r);
@@ -441,20 +575,20 @@ static int cgi_recv_response(request_st * const r, handler_ctx * const hctx) {
 			cgi_connection_close(hctx);
 			return HANDLER_FINISHED;
 		case HANDLER_COMEBACK:
-			/* flag for mod_cgi_handle_subrequest() */
-			hctx->conf.local_redir = 2;
-			buffer_clear(hctx->response);
-			return HANDLER_COMEBACK;
+			return cgi_local_redir(r, hctx); /* HANDLER_COMEBACK */
 		}
 }
 
 
 static handler_t cgi_handle_fdevent(void *ctx, int revents) {
 	handler_ctx *hctx = ctx;
-	request_st * const r = hctx->r;
+	hctx->rd_revents |= revents;
+	joblist_append(hctx->con);
+	return HANDLER_FINISHED;
+}
 
-	joblist_append(r->con);
 
+static handler_t cgi_process_rd_revents(handler_ctx * const hctx, request_st * const r, int revents) {
 	if (revents & FDEVENT_IN) {
 		handler_t rc = cgi_recv_response(r, hctx); /*(might invalidate hctx)*/
 		if (rc != HANDLER_GO_ON) return rc;         /*(unless HANDLER_GO_ON)*/
@@ -477,7 +611,7 @@ static handler_t cgi_handle_fdevent(void *ctx, int revents) {
 			} while (rc == HANDLER_GO_ON);           /*(unless HANDLER_GO_ON)*/
 			r->conf.stream_response_body = flags;
 			return rc; /* HANDLER_FINISHED or HANDLER_COMEBACK or HANDLER_ERROR */
-		} else if (!buffer_string_is_empty(hctx->response)) {
+		} else if (!buffer_is_blank(hctx->response)) {
 			/* unfinished header package which is a body in reality */
 			r->resp_body_started = 1;
 			if (0 != http_chunk_append_buffer(r, hctx->response)) {
@@ -493,37 +627,33 @@ static handler_t cgi_handle_fdevent(void *ctx, int revents) {
 		return HANDLER_ERROR;
 	}
 
-	return HANDLER_FINISHED;
+	return HANDLER_GO_ON;
 }
 
 
+__attribute_cold__
+__attribute_noinline__
+static void cgi_env_offset_resize(env_accum *env) {
+    chunk_buffer_prepare_append(env->boffsets, env->boffsets->size*2);
+    env->offsets = (uintptr_t *)(void *)env->boffsets->ptr;
+    env->osize = env->boffsets->size/sizeof(*env->offsets);
+}
+
 static int cgi_env_add(void *venv, const char *key, size_t key_len, const char *val, size_t val_len) {
 	env_accum *env = venv;
-	char *dst;
 
 	if (!key || (!val && val_len)) return -1;
 
-	if (env->size - env->used < key_len + val_len + 2) {
-		if (0 == env->size) env->size = 4096;
-		do { env->size *= 2; } while (env->size - env->used < key_len + val_len + 2);
-		env->ptr = realloc(env->ptr, env->size);
-		force_assert(env->ptr);
-	}
+	if (__builtin_expect( (env->osize == env->oused), 0))
+		cgi_env_offset_resize(env);
+	env->offsets[env->oused++] = env->b->used-1;
 
-	dst = env->ptr + env->used;
+	char * const dst = buffer_extend(env->b, key_len + val_len + 2);
 	memcpy(dst, key, key_len);
 	dst[key_len] = '=';
 	if (val_len) memcpy(dst + key_len + 1, val, val_len);
 	dst[key_len + 1 + val_len] = '\0';
 
-	if (env->osize == env->oused) {
-		env->osize += 16;
-		env->offsets = realloc(env->offsets, env->osize * sizeof(*env->offsets));
-		force_assert(env->offsets);
-	}
-	env->offsets[env->oused++] = env->used;
-	env->used += key_len + val_len + 2;
-
 	return 0;
 }
 
@@ -535,13 +665,12 @@ static int cgi_write_request(handler_ctx *hctx, int fd) {
 	chunkqueue_remove_finished_chunks(cq); /* unnecessary? */
 
 	/* old comment: windows doesn't support select() on pipes - wouldn't be easy to fix for all platforms.
-	 * solution: if this is still a problem on windows, then substitute
-	 * socketpair() for pipe() and closesocket() for close() on windows.
 	 */
 
 	for (c = cq->first; c; c = cq->first) {
 		ssize_t wr = chunkqueue_write_chunk_to_pipe(fd, cq, r->conf.errh);
 		if (wr > 0) {
+			hctx->write_ts = log_monotonic_secs;
 			chunkqueue_mark_written(cq, wr);
 			/* continue if wrote whole chunk or wrote 16k block
 			 * (see chunkqueue_write_chunk_file_intermed()) */
@@ -604,6 +733,7 @@ static int cgi_write_request(handler_ctx *hctx, int fd) {
 			}
 		} else {
 			/* more request body remains to be sent to CGI so register for fdevents */
+			hctx->write_ts = log_monotonic_secs;
 			fdevent_fdnode_event_set(ev, hctx->fdntocgi, FDEVENT_OUT);
 		}
 	}
@@ -615,10 +745,9 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 	char *args[3];
 	int to_cgi_fds[2];
 	int from_cgi_fds[2];
-	int dfd = -1;
 	UNUSED(p);
 
-	if (!buffer_string_is_empty(cgi_handler)) {
+	if (!buffer_is_blank(cgi_handler)) {
 		if (NULL == stat_cache_path_stat(cgi_handler)) {
 			log_perror(r->conf.errh, __FILE__, __LINE__,
 			  "stat for cgi-handler %s", cgi_handler->ptr);
@@ -626,22 +755,71 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 		}
 	}
 
-	if (pipe_cloexec(to_cgi_fds)) {
+	to_cgi_fds[0] = -1;
+  #ifndef __CYGWIN__
+	if (0 == r->reqbody_length) {
+		/* future: might keep fd open in p->devnull for reuse
+		 * and dup() here, or do not close() (later in this func) */
+		to_cgi_fds[0] = fdevent_open_devnull();
+		if (-1 == to_cgi_fds[0]) {
+			log_perror(r->conf.errh, __FILE__, __LINE__, "open /dev/null");
+			return -1;
+		}
+	}
+	else if (!(r->conf.stream_request_body /*(if not streaming request body)*/
+	           & (FDEVENT_STREAM_REQUEST|FDEVENT_STREAM_REQUEST_BUFMIN))) {
+		chunkqueue * const cq = &r->reqbody_queue;
+		chunk * const c = cq->first;
+		if (c && c == cq->last && c->type == FILE_CHUNK && c->file.is_temp) {
+			/* request body in single tempfile if not streaming req body */
+			if (-1 == c->file.fd
+			    && 0 != chunkqueue_open_file_chunk(cq, r->conf.errh))
+				return -1;
+		  #ifdef __COVERITY__
+			force_assert(-1 != c->file.fd);
+		  #endif
+			if (-1 == lseek(c->file.fd, 0, SEEK_SET)) {
+				log_perror(r->conf.errh, __FILE__, __LINE__,
+				  "lseek %s", c->mem->ptr);
+				return -1;
+			}
+			to_cgi_fds[0] = c->file.fd;
+			to_cgi_fds[1] = -1;
+		}
+	}
+  #endif
+
+	unsigned int bufsz_hint = 16384;
+  #ifdef _WIN32
+	if (r->reqbody_length <= 1048576)
+		bufsz_hint = (unsigned int)r->reqbody_length;
+  #endif
+	if (-1 == to_cgi_fds[0] && fdevent_pipe_cloexec(to_cgi_fds, bufsz_hint)) {
 		log_perror(r->conf.errh, __FILE__, __LINE__, "pipe failed");
 		return -1;
 	}
-	if (pipe_cloexec(from_cgi_fds)) {
-		close(to_cgi_fds[0]);
-		close(to_cgi_fds[1]);
+	if (fdevent_pipe_cloexec(from_cgi_fds, bufsz_hint)) {
+		if (0 == r->reqbody_length) {
+			close(to_cgi_fds[0]);
+		}
+		else if (-1 != to_cgi_fds[1]) {
+			close(to_cgi_fds[0]);
+			close(to_cgi_fds[1]);
+		}
 		log_perror(r->conf.errh, __FILE__, __LINE__, "pipe failed");
 		return -1;
 	}
 
+	env_accum * const env = &p->env;
+	env->b = chunk_buffer_acquire();
+	env->boffsets = chunk_buffer_acquire();
+	buffer_truncate(env->b, 0);
+	char **envp;
 	{
 		size_t i = 0;
 		http_cgi_opts opts = { 0, 0, NULL, NULL };
-		env_accum *env = &p->env;
-		env->used = 0;
+		env->offsets = (uintptr_t *)(void *)env->boffsets->ptr;
+		env->osize = env->boffsets->size/sizeof(*env->offsets);
 		env->oused = 0;
 
 		/* create environment */
@@ -650,72 +828,90 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 
 		/* for valgrind */
 		if (p->env.ld_preload) {
-			cgi_env_add(env, CONST_STR_LEN("LD_PRELOAD"), CONST_BUF_LEN(p->env.ld_preload));
+			cgi_env_add(env, CONST_STR_LEN("LD_PRELOAD"), BUF_PTR_LEN(p->env.ld_preload));
 		}
 		if (p->env.ld_library_path) {
-			cgi_env_add(env, CONST_STR_LEN("LD_LIBRARY_PATH"), CONST_BUF_LEN(p->env.ld_library_path));
+			cgi_env_add(env, CONST_STR_LEN("LD_LIBRARY_PATH"), BUF_PTR_LEN(p->env.ld_library_path));
 		}
 	      #ifdef __CYGWIN__
 		/* CYGWIN needs SYSTEMROOT */
 		if (p->env.systemroot) {
-			cgi_env_add(env, CONST_STR_LEN("SYSTEMROOT"), CONST_BUF_LEN(p->env.systemroot));
+			cgi_env_add(env, CONST_STR_LEN("SYSTEMROOT"), BUF_PTR_LEN(p->env.systemroot));
 		}
 	      #endif
 
-		if (env->esize <= env->oused) {
-			env->esize = (env->oused + 1 + 0xf) & ~(0xfuL);
-			env->eptr = realloc(env->eptr, env->esize * sizeof(*env->eptr));
-			force_assert(env->eptr);
-		}
-		for (i = 0; i < env->oused; ++i) {
-			env->eptr[i] = env->ptr + env->offsets[i];
-		}
-		env->eptr[env->oused] = NULL;
+		/* adjust (uintptr_t) offsets to (char *) ptr
+		 * (stored as offsets while accumulating in buffer,
+		 *  in case buffer is reallocated during env creation) */
+		if (__builtin_expect( (env->osize == env->oused), 0))
+			cgi_env_offset_resize(env);
+		envp = (char **)env->offsets;
+		envp[env->oused] = NULL;
+		const uintptr_t baseptr = (uintptr_t)env->b->ptr;
+		for (i = 0; i < env->oused; ++i)
+			envp[i] += baseptr;
 
 		/* set up args */
 		i = 0;
 
-		if (!buffer_string_is_empty(cgi_handler)) {
+		if (!buffer_is_blank(cgi_handler)) {
 			args[i++] = cgi_handler->ptr;
 		}
 		args[i++] = r->physical.path.ptr;
 		args[i  ] = NULL;
 	}
 
-	dfd = fdevent_open_dirname(r->physical.path.ptr, r->conf.follow_symlink);
+	int dfd = fdevent_open_dirname(r->physical.path.ptr,r->conf.follow_symlink);
 	if (-1 == dfd) {
 		log_perror(r->conf.errh, __FILE__, __LINE__, "open dirname %s failed", r->physical.path.ptr);
 	}
 
-	int serrh_fd = r->conf.serrh ? r->conf.serrh->errorlog_fd : -1;
-	hctx->pid = (dfd >= 0) ? fdevent_fork_execve(args[0], args, p->env.eptr, to_cgi_fds[0], from_cgi_fds[1], serrh_fd, dfd) : -1;
+	int serrh_fd = r->conf.serrh ? r->conf.serrh->fd : -1;
+	pid_t pid = (dfd >= 0)
+	  ? fdevent_fork_execve(args[0], args, envp,
+	                        to_cgi_fds[0], from_cgi_fds[1], serrh_fd, dfd)
+	  : -1;
 
-	if (-1 == hctx->pid) {
+	chunk_buffer_release(env->boffsets);
+	chunk_buffer_release(env->b);
+	env->boffsets = NULL;
+	env->b = NULL;
+
+	if (-1 == pid) {
 		/* log error with errno prior to calling close() (might change errno) */
 		log_perror(r->conf.errh, __FILE__, __LINE__, "fork failed");
 		if (-1 != dfd) close(dfd);
 		close(from_cgi_fds[0]);
 		close(from_cgi_fds[1]);
-		close(to_cgi_fds[0]);
-		close(to_cgi_fds[1]);
+		if (0 == r->reqbody_length) {
+			close(to_cgi_fds[0]);
+		}
+		else if (-1 != to_cgi_fds[1]) {
+			close(to_cgi_fds[0]);
+			close(to_cgi_fds[1]);
+		}
 		return -1;
 	} else {
 		if (-1 != dfd) close(dfd);
 		close(from_cgi_fds[1]);
-		close(to_cgi_fds[0]);
 
 		hctx->fd = from_cgi_fds[0];
-
-		cgi_pid_add(p, hctx->pid, hctx);
+		hctx->cgi_pid = cgi_pid_add(p, pid, hctx);
 
 		if (0 == r->reqbody_length) {
-			close(to_cgi_fds[1]);
+			close(to_cgi_fds[0]);
+		}
+		else if (-1 == to_cgi_fds[1]) {
+			chunkqueue * const cq = &r->reqbody_queue;
+			chunkqueue_mark_written(cq, chunkqueue_length(cq));
 		}
 		else if (0 == fdevent_fcntl_set_nb(to_cgi_fds[1])
 		         && 0 == cgi_write_request(hctx, to_cgi_fds[1])) {
+			close(to_cgi_fds[0]);
 			++r->con->srv->cur_fds;
 		}
 		else {
+			close(to_cgi_fds[0]);
 			close(to_cgi_fds[1]);
 			/*(hctx->fd not yet registered with fdevent, so manually
 			 * cleanup here; see fdevent_register() further below)*/
@@ -734,6 +930,7 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 			cgi_connection_close(hctx);
 			return -1;
 		}
+		hctx->read_ts = log_monotonic_secs;
 		fdevent_fdnode_event_set(ev, hctx->fdn, FDEVENT_IN | FDEVENT_RDHUP);
 
 		return 0;
@@ -746,7 +943,8 @@ URIHANDLER_FUNC(cgi_is_handled) {
 	data_string *ds;
 
 	if (NULL != r->handler_module) return HANDLER_GO_ON;
-	if (buffer_is_empty(&r->physical.path)) return HANDLER_GO_ON;
+	/* r->physical.path is non-empty for handle_subrequest_start */
+	/*if (buffer_is_blank(&r->physical.path)) return HANDLER_GO_ON;*/
 
 	mod_cgi_patch_config(r, p);
 	if (NULL == p->conf.cgi) return HANDLER_GO_ON;
@@ -754,24 +952,48 @@ URIHANDLER_FUNC(cgi_is_handled) {
 	ds = (data_string *)array_match_key_suffix(p->conf.cgi, &r->physical.path);
 	if (NULL == ds) return HANDLER_GO_ON;
 
-	st = stat_cache_path_stat(&r->physical.path);
+	/* r->tmp_sce is set in http_response_physical_path_check() and is valid
+	 * in handle_subrequest_start callback -- handle_subrequest_start callbacks
+	 * should not change r->physical.path (or should invalidate r->tmp_sce) */
+	st = r->tmp_sce && buffer_is_equal(&r->tmp_sce->name, &r->physical.path)
+	   ? &r->tmp_sce->st
+	   : stat_cache_path_stat(&r->physical.path);
 	if (NULL == st) return HANDLER_GO_ON;
 
 	/* (aside: CGI might be executable even if it is not readable) */
 	if (!S_ISREG(st->st_mode)) return HANDLER_GO_ON;
 	if (p->conf.execute_x_only == 1 && (st->st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) == 0) return HANDLER_GO_ON;
 
+	if (r->reqbody_length
+	    && p->tempfile_accum
+	    && !(r->conf.stream_request_body /*(if not streaming request body)*/
+	         & (FDEVENT_STREAM_REQUEST|FDEVENT_STREAM_REQUEST_BUFMIN))) {
+		/* store request body in single tempfile if not streaming request body*/
+		r->reqbody_queue.upload_temp_file_size = INTMAX_MAX;
+	}
+
 	{
 		handler_ctx *hctx = cgi_handler_ctx_init();
 		hctx->ev = r->con->srv->ev;
 		hctx->r = r;
+		hctx->con = r->con;
 		hctx->plugin_data = p;
 		hctx->cgi_handler = &ds->value;
 		memcpy(&hctx->conf, &p->conf, sizeof(plugin_config));
-		hctx->conf.upgrade =
-		  hctx->conf.upgrade
-		  && r->http_version == HTTP_VERSION_1_1
-		  && light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE);
+		if (!light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE))
+			hctx->conf.upgrade = 0;
+		else if (!hctx->conf.upgrade || r->http_version != HTTP_VERSION_1_1) {
+			hctx->conf.upgrade = 0;
+			http_header_request_unset(r, HTTP_HEADER_UPGRADE,
+			                          CONST_STR_LEN("Upgrade"));
+		}
+		hctx->opts.max_per_read =
+		  !(r->conf.stream_response_body /*(if not streaming response body)*/
+		    & (FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN))
+		    ? 262144
+		    : (r->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN)
+		      ? 16384  /* FDEVENT_STREAM_RESPONSE_BUFMIN */
+		      : 65536; /* FDEVENT_STREAM_RESPONSE */
 		hctx->opts.fdfmt = S_IFIFO;
 		hctx->opts.backend = BACKEND_CGI;
 		hctx->opts.authorizer = 0;
@@ -787,17 +1009,6 @@ URIHANDLER_FUNC(cgi_is_handled) {
 	return HANDLER_GO_ON;
 }
 
-__attribute_cold__
-__attribute_noinline__
-static handler_t mod_cgi_local_redir(request_st * const r) {
-    /* must be called from mod_cgi_handle_subrequest() so that HANDLER_COMEBACK
-     * return value propagates back through connection_state_machine() */
-    http_response_reset(r); /*(includes r->http_status = 0)*/
-    plugins_call_handle_request_reset(r);
-    /*cgi_connection_close(hctx);*//*(already cleaned up and hctx is now invalid)*/
-    return HANDLER_COMEBACK;
-}
-
 /*
  * - HANDLER_GO_ON : not our job
  * - HANDLER_FINISHED: got response
@@ -808,7 +1019,28 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 	handler_ctx * const hctx = r->plugin_ctx[p->id];
 	if (NULL == hctx) return HANDLER_GO_ON;
 
-	if (2 == hctx->conf.local_redir) return mod_cgi_local_redir(r);
+	if (__builtin_expect(
+	     (r->conf.stream_request_body & FDEVENT_STREAM_REQUEST_TCP_FIN), 0)
+	    && hctx->conf.limits && hctx->conf.limits->signal_fin) {
+		/* XXX: consider setting r->http_status = 499 if (0 == r->http_status)
+		 * (499 is nginx custom status to indicate client closed connection) */
+		if (-1 == hctx->fd) return HANDLER_ERROR; /*(CGI not yet spawned)*/
+		if (hctx->cgi_pid) /* send signal to notify CGI about TCP FIN */
+			cgi_pid_kill(hctx->cgi_pid, hctx->conf.limits->signal_fin);
+	}
+
+	const int rd_revents = hctx->rd_revents;
+	const int wr_revents = hctx->wr_revents;
+	if (rd_revents) {
+		hctx->rd_revents = 0;
+		handler_t rc = cgi_process_rd_revents(hctx, r, rd_revents);
+		if (rc != HANDLER_GO_ON) return rc; /*(might invalidate hctx)*/
+	}
+	if (wr_revents) {
+		hctx->wr_revents = 0;
+		handler_t rc = cgi_process_wr_revents(hctx, r, wr_revents);
+		if (rc != HANDLER_GO_ON) return rc; /*(might invalidate hctx)*/
+	}
 
 	if ((r->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN)
 	    && r->resp_body_started) {
@@ -817,8 +1049,8 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 		} else if (!(fdevent_fdnode_interest(hctx->fdn) & FDEVENT_IN)) {
 			/* optimistic read from backend */
 			handler_t rc = cgi_recv_response(r, hctx);  /*(might invalidate hctx)*/
-			if (rc == HANDLER_COMEBACK) mod_cgi_local_redir(r);
 			if (rc != HANDLER_GO_ON) return rc;          /*(unless HANDLER_GO_ON)*/
+			hctx->read_ts = log_monotonic_secs;
 			fdevent_fdnode_event_add(hctx->ev, hctx->fdn, FDEVENT_IN);
 		}
 	}
@@ -872,22 +1104,73 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 }
 
 
-static handler_t cgi_waitpid_cb(server *srv, void *p_d, pid_t pid, int status) {
-    plugin_data *p = (plugin_data *)p_d;
-    for (size_t i = 0; i < p->cgi_pid.used; ++i) {
-        handler_ctx *hctx;
-        if (pid != p->cgi_pid.ptr[i].pid) continue;
+__attribute_cold__
+__attribute_noinline__
+static void cgi_trigger_hctx_timeout(handler_ctx * const hctx, const char * const msg) {
+    request_st * const r = hctx->r;
+    joblist_append(r->con);
 
-        hctx = (handler_ctx *)p->cgi_pid.ptr[i].ctx;
-        if (hctx) hctx->pid = -1;
-        cgi_pid_del(p, i);
+    log_error(r->conf.errh, __FILE__, __LINE__,
+      "%s timeout on CGI: %s (pid: %lld)",
+      msg, r->physical.path.ptr, (long long)hctx->cgi_pid->pid);
+
+    if (*msg == 'w') { /* "write" */
+        /* theoretically, response might be waiting on hctx->fdn pipe
+         * if it arrived since we last checked for event, and if CGI
+         * timeout out while reading (or did not read) request body */
+        handler_t rc = cgi_recv_response(r, hctx); /*(might invalidate hctx)*/
+        if (rc != HANDLER_GO_ON) return;            /*(unless HANDLER_GO_ON)*/
+    }
+
+    if (0 == r->http_status) r->http_status = 504; /* Gateway Timeout */
+    cgi_connection_close(hctx);
+}
+
+
+static handler_t cgi_trigger_cb(server *srv, void *p_d) {
+    UNUSED(srv);
+    const unix_time64_t mono = log_monotonic_secs;
+    plugin_data * const p = p_d;
+    for (cgi_pid_t *cgi_pid = p->cgi_pid; cgi_pid; cgi_pid = cgi_pid->next) {
+        /*(hctx stays in cgi_pid list until process pid is reaped,
+         * so p->cgi_pid[] is not modified during this loop)*/
+        handler_ctx * const hctx = cgi_pid->hctx;
+        if (!hctx) continue; /*(already called cgi_pid_kill())*/
+        const cgi_limits * const limits = hctx->conf.limits;
+        if (NULL == limits) continue;
+        if (limits->read_timeout && hctx->fdn
+            && (fdevent_fdnode_interest(hctx->fdn) & FDEVENT_IN)
+            && mono - hctx->read_ts > limits->read_timeout) {
+            cgi_trigger_hctx_timeout(hctx, "read");
+            continue;
+        }
+        if (limits->write_timeout && hctx->fdntocgi
+            && (fdevent_fdnode_interest(hctx->fdntocgi) & FDEVENT_OUT)
+            && mono - hctx->write_ts > limits->write_timeout) {
+            cgi_trigger_hctx_timeout(hctx, "write");
+            continue;
+        }
+    }
+    return HANDLER_GO_ON;
+}
+
+
+static handler_t cgi_waitpid_cb(server *srv, void *p_d, pid_t pid, int status) {
+    /*(XXX: if supporting a large number of CGI, might use a different algorithm
+     * instead of linked list, e.g. splaytree indexed with pid)*/
+    plugin_data *p = (plugin_data *)p_d;
+    for (cgi_pid_t *cgi_pid = p->cgi_pid; cgi_pid; cgi_pid = cgi_pid->next) {
+        if (pid != cgi_pid->pid) continue;
+
+        handler_ctx * const hctx = cgi_pid->hctx;
+        if (hctx) hctx->cgi_pid = NULL;
 
         if (WIFEXITED(status)) {
             /* (skip logging (non-zero) CGI exit; might be very noisy) */
         }
         else if (WIFSIGNALED(status)) {
             /* ignore SIGTERM if sent by cgi_connection_close() (NULL == hctx)*/
-            if (WTERMSIG(status) != SIGTERM || NULL != hctx) {
+            if (WTERMSIG(status) != cgi_pid->signal_sent) {
                 log_error_st *errh = hctx ? hctx->r->conf.errh : srv->errh;
                 log_error(errh, __FILE__, __LINE__,
                   "CGI pid %d died with signal %d", pid, WTERMSIG(status));
@@ -899,6 +1182,7 @@ static handler_t cgi_waitpid_cb(server *srv, void *p_d, pid_t pid, int status) {
               "CGI pid %d ended unexpectedly", pid);
         }
 
+        cgi_pid_del(p, cgi_pid);
         return HANDLER_FINISHED;
     }
 
@@ -914,6 +1198,7 @@ int mod_cgi_plugin_init(plugin *p) {
 	p->handle_request_reset = cgi_connection_close_callback;
 	p->handle_subrequest_start = cgi_is_handled;
 	p->handle_subrequest = mod_cgi_handle_subrequest;
+	p->handle_trigger = cgi_trigger_cb;
 	p->handle_waitpid = cgi_waitpid_cb;
 	p->init           = mod_cgi_init;
 	p->cleanup        = mod_cgi_free;
