@@ -703,8 +703,8 @@ static const char js_simple_table_resort[] = \
 " var bt = get_inner_text(b.cells[sort_column]);\n" \
 " var cmp;\n" \
 " if (sort_column == name_column) {\n" \
-"  if (at == '..') return -1;\n" \
-"  if (bt == '..') return  1;\n" \
+"  if (at == '../') return -1;\n" \
+"  if (bt == '../') return  1;\n" \
 " }\n" \
 " if (a.cells[sort_column].className == 'int') {\n" \
 "  cmp = parseInt(at)-parseInt(bt);\n" \
@@ -1077,6 +1077,7 @@ static int http_read_directory(handler_ctx * const p) {
 		}
 		dirls_entry_t * const tmp = list->ent[list->used++] =
 		  (dirls_entry_t*) malloc(sizeof(dirls_entry_t) + 1 + dsz);
+		force_assert(tmp);
 		tmp->mtime = st.st_mtime;
 		tmp->size  = st.st_size;
 		tmp->namelen = dsz;
