@@ -93,6 +93,25 @@ SETDEFAULTS_FUNC(mod_evasive_set_defaults) {
     if (!config_plugin_values_init(srv, p, cpk, "mod_evasive"))
         return HANDLER_ERROR;
 
+    /* process and validate config directives
+     * (init i to 0 if global context; to 1 to skip empty global context) */
+    for (int i = !p->cvlist[0].v.u2[1]; i < p->nconfig; ++i) {
+        config_plugin_value_t *cpv = p->cvlist + p->cvlist[i].v.u2[0];
+        for (; -1 != cpv->k_id; ++cpv) {
+            switch (cpv->k_id) {
+              case 0: /* evasive.max-conns-per-ip */
+              case 1: /* evasive.silent */
+                break;
+              case 2: /* evasive.location */
+                if (buffer_is_blank(cpv->v.b))
+                    cpv->v.b = NULL;
+                break;
+              default:/* should not happen */
+                break;
+            }
+        }
+    }
+
     /* initialize p->defaults from global config context */
     if (p->nconfig > 0 && p->cvlist->v.u2[1]) {
         const config_plugin_value_t *cpv = p->cvlist + p->cvlist->v.u2[0];
@@ -112,13 +131,11 @@ URIHANDLER_FUNC(mod_evasive_uri_handler) {
 	if (p->conf.max_conns == 0) return HANDLER_GO_ON;
 
 	sock_addr * const dst_addr = &r->con->dst_addr;
-	const connections * const conns = &r->con->srv->conns;
-	for (uint32_t i = 0, conns_by_ip = 0; i < conns->used; ++i) {
-		connection *c = conns->ptr[i];
-
+	uint32_t conns_by_ip = 0;
+	for (const connection *c = r->con->srv->conns; c; c = c->next) {
 		/* check if other connections are already actively serving data for the same IP
 		 * we can only ban connections which are already behind the 'read request' state
-		 * */
+		 */
 		if (c->request.state <= CON_STATE_REQUEST_END) continue;
 
 		if (!sock_addr_is_addr_eq(&c->dst_addr, dst_addr)) continue;
@@ -128,11 +145,13 @@ URIHANDLER_FUNC(mod_evasive_uri_handler) {
 			if (!p->conf.silent) {
 				log_error(r->conf.errh, __FILE__, __LINE__,
 				  "%s turned away. Too many connections.",
-				  r->con->dst_addr_buf->ptr);
+				  r->con->dst_addr_buf.ptr);
 			}
 
-			if (!buffer_is_empty(p->conf.location)) {
-				http_header_response_set(r, HTTP_HEADER_LOCATION, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->conf.location));
+			if (p->conf.location) {
+				http_header_response_set(r, HTTP_HEADER_LOCATION,
+				                         CONST_STR_LEN("Location"),
+				                         BUF_PTR_LEN(p->conf.location));
 				r->http_status = 302;
 				r->resp_body_finished = 1;
 			} else {
