@@ -94,6 +94,7 @@ static void mod_rewrite_patch_config(request_st * const r, plugin_data * const p
 }
 
 static pcre_keyvalue_buffer * mod_rewrite_parse_list(server *srv, const array *a, pcre_keyvalue_buffer *kvb, const int condidx) {
+    const int pcre_jit = config_feature_bool(srv, "server.pcre_jit", 1);
     int allocated = 0;
     if (NULL == kvb) {
         allocated = 1;
@@ -108,7 +109,8 @@ static pcre_keyvalue_buffer * mod_rewrite_parse_list(server *srv, const array *a
             pcre_keyvalue_burl_normalize_key(&ds->key, tb);
             pcre_keyvalue_burl_normalize_value(&ds->value, tb);
         }
-        if (!pcre_keyvalue_buffer_append(srv->errh, kvb, &ds->key, &ds->value)){
+        if (!pcre_keyvalue_buffer_append(srv->errh, kvb, &ds->key, &ds->value,
+                                         pcre_jit)) {
             log_error(srv->errh, __FILE__, __LINE__,
               "pcre-compile failed for %s", ds->key.ptr);
             if (allocated) pcre_keyvalue_buffer_free(kvb);
@@ -285,12 +287,12 @@ static handler_t process_rewrite_rules(request_st * const r, plugin_data *p, con
 	burl.port      = sock_addr_get_port(&r->con->srv_socket->addr);
 	burl.path      = &r->target; /*(uri-encoded and includes query-part)*/
 	burl.query     = &r->uri.query;
-	if (buffer_string_is_empty(burl.authority))
+	if (buffer_is_blank(burl.authority))
 		burl.authority = r->server_name;
 
 	buffer * const tb = r->tmp_buf;
 	rc = pcre_keyvalue_buffer_process(kvb, &ctx, &r->target, tb);
-	if (HANDLER_FINISHED == rc && !buffer_is_empty(tb) && tb->ptr[0] == '/') {
+	if (HANDLER_FINISHED == rc && !buffer_is_blank(tb) && tb->ptr[0] == '/') {
 		buffer_copy_buffer(&r->target, tb);
 		uintptr_t * const hctx = (uintptr_t *)(r->plugin_ctx + p->id);
 		*hctx |= REWRITE_STATE_REWRITTEN;
