@@ -5,6 +5,7 @@
 #include "base_decls.h"
 #include "buffer.h"
 #include "array.h"
+#include "chunk.h"
 
 #include "plugin.h"
 
@@ -20,27 +21,23 @@ typedef struct {
 	PLUGIN_DATA;
 	plugin_config defaults;
 	plugin_config conf;
-
-	buffer *timefmt;
-
-	buffer *stat_fn;
-
 	array *ssi_vars;
 	array *ssi_cgi_env;
+	buffer stat_fn;
+	buffer timefmt;
 } plugin_data;
 
 typedef struct {
-	buffer *timefmt;
-	int sizefmt;
-
-	buffer *stat_fn;
-
 	array *ssi_vars;
 	array *ssi_cgi_env;
+	buffer *stat_fn;
+	buffer *timefmt;
+	int sizefmt;
 
 	int if_level, if_is_false_level, if_is_false, if_is_false_endif;
 	unsigned short ssi_recursion_depth;
 
+	chunkqueue wq;
 	log_error_st *errh;
 	plugin_config conf;
 } handler_ctx;
