@@ -3,7 +3,6 @@
 #include "plugin.h"
 
 
-#if defined(HAVE_PCRE_H)                /* do nothing if PCRE not available */
 #if defined(HAVE_GDBM_H) || defined(USE_MEMCACHED) /* at least one required */
 
 
@@ -11,6 +10,7 @@
 #include "log.h"
 #include "buffer.h"
 #include "http_header.h"
+#include "keyvalue.h"
 
 #include <sys/stat.h>
 #include <stdlib.h>
@@ -21,10 +21,6 @@
 # include <gdbm.h>
 #endif
 
-#if defined(HAVE_PCRE_H)
-# include <pcre.h>
-#endif
-
 #if defined(USE_MEMCACHED)
 # include <libmemcached/memcached.h>
 #endif
@@ -36,8 +32,8 @@
 
 typedef struct {
     const buffer *deny_url;
-    pcre *trigger_regex;
-    pcre *download_regex;
+    pcre_keyvalue_buffer *trigger_regex;
+    pcre_keyvalue_buffer *download_regex;
   #if defined(HAVE_GDBM_H)
     GDBM_FILE db;
   #endif
@@ -74,10 +70,10 @@ FREE_FUNC(mod_trigger_b4_dl_free) {
                 break;
              #endif
               case 1: /* trigger-before-download.trigger-url */
-                pcre_free(cpv->v.v);
+                pcre_keyvalue_buffer_free(cpv->v.v);
                 break;
               case 2: /* trigger-before-download.download-url */
-                pcre_free(cpv->v.v);
+                pcre_keyvalue_buffer_free(cpv->v.v);
                 break;
              #if defined(USE_MEMCACHED)
               case 5: /* trigger-before-download.memcache-hosts */
@@ -92,7 +88,7 @@ FREE_FUNC(mod_trigger_b4_dl_free) {
 }
 
 static int mod_trigger_b4_dl_init_gdbm(server * const srv, config_plugin_value_t * const cpv) {
-    if (buffer_string_is_empty(cpv->v.b)) {
+    if (buffer_is_blank(cpv->v.b)) {
         cpv->v.v = NULL;
         return 1;
     }
@@ -135,11 +131,11 @@ static int mod_trigger_b4_dl_init_memcached(server * const srv, config_plugin_va
     buffer_clear(opts);
     for (uint32_t k = 0; k < mc_hosts->used; ++k) {
         const data_string * const ds = (const data_string *)mc_hosts->data[k];
-        buffer_append_string_len(opts, CONST_STR_LEN(" --SERVER="));
-        buffer_append_string_buffer(opts, &ds->value);
+        buffer_append_str2(opts, CONST_STR_LEN(" --SERVER="),
+                                 BUF_PTR_LEN(&ds->value));
     }
 
-    cpv->v.v = memcached(opts->ptr+1, buffer_string_length(opts)-1);
+    cpv->v.v = memcached(opts->ptr+1, buffer_clen(opts)-1);
 
     if (cpv->v.v) {
         cpv->vtype = T_CONFIG_LOCAL;
@@ -163,25 +159,44 @@ static int mod_trigger_b4_dl_init_memcached(server * const srv, config_plugin_va
 
 static int mod_trigger_b4_dl_init_regex(server * const srv, config_plugin_value_t * const cpv, const char * const str) {
     const buffer * const b = cpv->v.b;
-    if (buffer_string_is_empty(b)) {
+    if (buffer_is_blank(b)) {
         cpv->v.v = NULL;
         return 1;
     }
 
-    const char *errptr;
-    int erroff;
-    cpv->v.v = pcre_compile(b->ptr, 0, &errptr, &erroff, NULL);
-
-    if (cpv->v.v) {
-        cpv->vtype = T_CONFIG_LOCAL;
-        return 1;
-    }
-    else {
+    const int pcre_jit = config_feature_bool(srv, "server.pcre_jit", 1);
+    pcre_keyvalue_buffer * const kvb = pcre_keyvalue_buffer_init();
+    buffer empty = { NULL, 0, 0 };
+    if (!pcre_keyvalue_buffer_append(srv->errh, kvb, b, &empty, pcre_jit)) {
         log_error(srv->errh, __FILE__, __LINE__,
-          "compiling regex for %s failed: %s pos: %d",
-          str, b->ptr, erroff);
+          "pcre_compile failed for %s %s", str, b->ptr);
+        pcre_keyvalue_buffer_free(kvb);
         return 0;
     }
+    cpv->v.v = kvb;
+    cpv->vtype = T_CONFIG_LOCAL;
+    return 1;
+}
+
+#ifdef __COVERITY__
+#include "burl.h"
+#endif
+
+static int mod_trigger_b4_dl_match(pcre_keyvalue_buffer * const kvb, const buffer * const input) {
+    /*(re-use keyvalue.[ch] for match-only;
+     *  must have been configured with empty kvb 'value' during init)*/
+    pcre_keyvalue_ctx ctx = { NULL, NULL, 0, -1 };
+  #ifdef __COVERITY__
+    /*(again, must have been configured w/ empty kvb 'value' during init)*/
+    struct cond_match_t cache;
+    memset(&cache, 0, sizeof(cache));
+    struct burl_parts_t bp;
+    memset(&bp, 0, sizeof(bp));
+    ctx.cache = &cache;
+    ctx.burl = &bp;
+  #endif
+    return HANDLER_GO_ON == pcre_keyvalue_buffer_process(kvb, &ctx, input, NULL)
+        && -1 != ctx.m;
 }
 
 static void mod_trigger_b4_dl_merge_config_cpv(plugin_config * const pconf, const config_plugin_value_t * const cpv) {
@@ -327,7 +342,7 @@ static void mod_trigger_b4_dl_memcached_key(buffer * const b, const plugin_data
     buffer_append_string_buffer(b, remote_ip);
 
     /* memcached can't handle spaces */
-    for (size_t i = 0, len = buffer_string_length(b); i < len; ++i) {
+    for (size_t i = 0, len = buffer_clen(b); i < len; ++i) {
         if (b->ptr[i] == ' ') b->ptr[i] = '-';
     }
 }
@@ -337,7 +352,7 @@ static handler_t mod_trigger_b4_dl_deny(request_st * const r, const plugin_data
     if (p->conf.deny_url) {
         http_header_response_set(r, HTTP_HEADER_LOCATION,
                                  CONST_STR_LEN("Location"),
-                                 CONST_BUF_LEN(p->conf.deny_url));
+                                 BUF_PTR_LEN(p->conf.deny_url));
         r->http_status = 307;
     }
     else {
@@ -352,10 +367,6 @@ static handler_t mod_trigger_b4_dl_deny(request_st * const r, const plugin_data
 URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	plugin_data *p = p_d;
 
-	int n;
-# define N 10
-	int ovec[N * 3];
-
 	if (NULL != r->handler_module) return HANDLER_GO_ON;
 
 	mod_trigger_b4_dl_patch_config(r, p);
@@ -382,31 +393,24 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	  http_header_request_get(r, HTTP_HEADER_X_FORWARDED_FOR,
 	                          CONST_STR_LEN("X-Forwarded-For"));
 	if (NULL == remote_ip) {
-		remote_ip = r->con->dst_addr_buf;
+		remote_ip = &r->con->dst_addr_buf;
 	}
 
 	if (p->conf.debug) {
 		log_error(r->conf.errh, __FILE__, __LINE__, "(debug) remote-ip: %s", remote_ip->ptr);
 	}
 
-	const time_t cur_ts = log_epoch_secs;
+	const unix_time64_t cur_ts = log_epoch_secs;
 
 	/* check if URL is a trigger -> insert IP into DB */
-	if ((n = pcre_exec(p->conf.trigger_regex, NULL, CONST_BUF_LEN(&r->uri.path), 0, 0, ovec, 3 * N)) < 0) {
-		if (n != PCRE_ERROR_NOMATCH) {
-			log_error(r->conf.errh, __FILE__, __LINE__,
-			  "execution error while matching: %d", n);
-
-			return HANDLER_ERROR;
-		}
-	} else {
+	if (mod_trigger_b4_dl_match(p->conf.trigger_regex, &r->uri.path)) {
+		/* the trigger matched */
 # if defined(HAVE_GDBM_H)
 		if (p->conf.db) {
-			/* the trigger matched */
 			datum key, val;
 
 			*(const char **)&key.dptr = remote_ip->ptr;
-			key.dsize = buffer_string_length(remote_ip);
+			key.dsize = buffer_clen(remote_ip);
 
 			val.dptr = (char *)&cur_ts;
 			val.dsize = sizeof(cur_ts);
@@ -426,7 +430,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 			}
 
 			if (MEMCACHED_SUCCESS != memcached_set(p->conf.memc,
-					CONST_BUF_LEN(b),
+					BUF_PTR_LEN(b),
 					(const char *)&cur_ts, sizeof(cur_ts),
 					p->conf.trigger_timeout, 0)) {
 				log_error(r->conf.errh, __FILE__, __LINE__, "insert failed");
@@ -436,21 +440,15 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	}
 
 	/* check if URL is a download -> check IP in DB, update timestamp */
-	if ((n = pcre_exec(p->conf.download_regex, NULL, CONST_BUF_LEN(&r->uri.path), 0, 0, ovec, 3 * N)) < 0) {
-		if (n != PCRE_ERROR_NOMATCH) {
-			log_error(r->conf.errh, __FILE__, __LINE__,
-			  "execution error while matching: %d", n);
-			return HANDLER_ERROR;
-		}
-	} else {
+	if (mod_trigger_b4_dl_match(p->conf.download_regex, &r->uri.path)) {
 		/* the download uri matched */
 # if defined(HAVE_GDBM_H)
 		if (p->conf.db) {
 			datum key, val;
-			time_t last_hit;
+			unix_time64_t last_hit = 0;
 
 			*(const char **)&key.dptr = remote_ip->ptr;
-			key.dsize = buffer_string_length(remote_ip);
+			key.dsize = buffer_clen(remote_ip);
 
 			val = gdbm_fetch(p->conf.db, key);
 
@@ -459,7 +457,13 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 				return mod_trigger_b4_dl_deny(r, p);
 			}
 
-			memcpy(&last_hit, val.dptr, sizeof(time_t));
+			if (val.dsize == sizeof(last_hit))
+				memcpy(&last_hit, val.dptr, val.dsize);
+			else if (val.dsize == 4) {
+				int32_t t;
+				memcpy(&t, val.dptr, val.dsize);
+				last_hit = t;
+			}
 
 			free(val.dptr);
 
@@ -498,13 +502,13 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 			 * and the timestamp is updated
 			 *
 			 */
-			if (MEMCACHED_SUCCESS != memcached_exist(p->conf.memc, CONST_BUF_LEN(b))) {
+			if (MEMCACHED_SUCCESS != memcached_exist(p->conf.memc, BUF_PTR_LEN(b))) {
 				return mod_trigger_b4_dl_deny(r, p);
 			}
 
 			/* set a new timeout */
 			if (MEMCACHED_SUCCESS != memcached_set(p->conf.memc,
-					CONST_BUF_LEN(b),
+					BUF_PTR_LEN(b),
 					(const char *)&cur_ts, sizeof(cur_ts),
 					p->conf.trigger_timeout, 0)) {
 				log_error(r->conf.errh, __FILE__, __LINE__, "insert failed");
@@ -517,7 +521,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 }
 
 #if defined(HAVE_GDBM_H)
-static void mod_trigger_b4_dl_trigger_gdbm(GDBM_FILE db, const time_t cur_ts, const int trigger_timeout) {
+static void mod_trigger_b4_dl_trigger_gdbm(GDBM_FILE db, const unix_time64_t cur_ts, const int trigger_timeout) {
 		datum key, val, okey;
 		okey.dptr = NULL;
 
@@ -526,7 +530,7 @@ static void mod_trigger_b4_dl_trigger_gdbm(GDBM_FILE db, const time_t cur_ts, co
 		 * we don't care as the next round will remove them. We don't have to perfect here.
 		 */
 		for (key = gdbm_firstkey(db); key.dptr; key = gdbm_nextkey(db, okey)) {
-			time_t last_hit;
+			unix_time64_t last_hit = 0;
 			if (okey.dptr) {
 				free(okey.dptr);
 				okey.dptr = NULL;
@@ -534,7 +538,13 @@ static void mod_trigger_b4_dl_trigger_gdbm(GDBM_FILE db, const time_t cur_ts, co
 
 			val = gdbm_fetch(db, key);
 
-			memcpy(&last_hit, val.dptr, sizeof(time_t));
+			if (val.dsize == sizeof(last_hit))
+				memcpy(&last_hit, val.dptr, val.dsize);
+			else if (val.dsize == 4) {
+				int32_t t;
+				memcpy(&t, val.dptr, val.dsize);
+				last_hit = t;
+			}
 
 			free(val.dptr);
 
@@ -547,12 +557,12 @@ static void mod_trigger_b4_dl_trigger_gdbm(GDBM_FILE db, const time_t cur_ts, co
 		if (okey.dptr) free(okey.dptr);
 
 		/* reorg once a day */
-		if ((cur_ts % (60 * 60 * 24) != 0)) gdbm_reorganize(db);
+		if ((cur_ts % (60 * 60 * 24) == 0)) gdbm_reorganize(db);
 }
 
 TRIGGER_FUNC(mod_trigger_b4_dl_handle_trigger) {
     /* check DB each minute */
-    const time_t cur_ts = log_epoch_secs;
+    const unix_time64_t cur_ts = log_epoch_secs;
     if (cur_ts % 60 != 0) return HANDLER_GO_ON;
     UNUSED(srv);
 
@@ -585,7 +595,6 @@ TRIGGER_FUNC(mod_trigger_b4_dl_handle_trigger) {
 #endif
 
 
-#endif /* defined(HAVE_PCRE_H) */
 #endif /* defined(HAVE_GDBM_H) || defined(USE_MEMCACHED) */
 
 
@@ -594,7 +603,6 @@ int mod_trigger_b4_dl_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = "trigger_b4_dl";
 
-#if defined(HAVE_PCRE_H)                /* do nothing if PCRE not available */
 #if defined(HAVE_GDBM_H) || defined(USE_MEMCACHED) /* at least one required */
 
 	p->init        = mod_trigger_b4_dl_init;
@@ -605,7 +613,6 @@ int mod_trigger_b4_dl_plugin_init(plugin *p) {
 #endif
 	p->cleanup     = mod_trigger_b4_dl_free;
 
-#endif
 #endif
 
 	return 0;
