@@ -8,6 +8,7 @@
 #include "reqpool.h"
 
 #include <stdlib.h>
+#include <string.h>
 
 #include "base.h"
 #include "buffer.h"
@@ -18,6 +19,24 @@
 #include "response.h"
 
 
+static const request_config *request_config_defaults;
+
+
+void
+request_config_set_defaults (const request_config *config_defaults)
+{
+    request_config_defaults = config_defaults;
+}
+
+
+__attribute_noinline__
+void
+request_config_reset (request_st * const r)
+{
+    memcpy(&r->conf, request_config_defaults, sizeof(request_config));
+}
+
+
 void
 request_init_data (request_st * const r, connection * const con, server * const srv)
 {
@@ -32,6 +51,7 @@ request_init_data (request_st * const r, connection * const con, server * const
     r->con = con;
     r->tmp_buf = srv->tmp_buf;
     r->resp_body_scratchpad = -1;
+    r->server_name = &r->uri.authority;
 
     /* init plugin-specific per-request structures */
     r->plugin_ctx = calloc(1, (srv->plugins.used + 1) * sizeof(void *));
@@ -47,6 +67,8 @@ request_init_data (request_st * const r, connection * const con, server * const
         force_assert(NULL != r->cond_match);
     }
   #endif
+
+    request_config_reset(r);
 }
 
 
@@ -58,6 +80,7 @@ request_reset (request_st * const r)
     http_response_reset(r);
 
     r->loops_per_request = 0;
+    r->keep_alive = 0;
 
     r->h2state = 0; /* H2_STATE_IDLE */
     r->h2id = 0;
@@ -118,6 +141,8 @@ request_reset (request_st * const r)
 
     /* The cond_cache gets reset in response.c */
     /* config_cond_cache_reset(r); */
+
+    request_config_reset(r);
 }
 
 
@@ -155,9 +180,12 @@ request_reset_ex (request_st * const r)
     }
   #endif
 
+    r->server_name = &r->uri.authority;
     buffer_clear(&r->uri.authority);
     buffer_reset(&r->uri.path);
     buffer_reset(&r->uri.query);
+    buffer_reset(&r->physical.path);
+    buffer_reset(&r->physical.rel_path);
     buffer_reset(&r->target_orig);
     buffer_reset(&r->target);       /*(see comments in request_reset())*/
     buffer_reset(&r->pathinfo);     /*(see comments in request_reset())*/
@@ -204,15 +232,6 @@ request_free_data (request_st * const r)
 
 /* linked list of (request_st *) cached for reuse */
 static request_st *reqpool;
-/* max num of (request_st *) to cache */
-static uint32_t reqspace;
-
-
-void
-request_pool_init (uint32_t sz)
-{
-    reqspace = sz;
-}
 
 
 void
@@ -223,7 +242,6 @@ request_pool_free (void)
         reqpool = (request_st *)r->con; /*(reuse r->con as next ptr)*/
         request_free_data(r);
         free(r);
-        ++reqspace;
     }
 }
 
@@ -245,15 +263,8 @@ request_release (request_st * const r)
     request_reset_ex(r);
     r->state = CON_STATE_CONNECT;
 
-    if (reqspace) {
-        --reqspace;
-        r->con = (connection *)reqpool; /*(reuse r->con as next ptr)*/
-        reqpool = r;
-    }
-    else {
-        request_free_data(r);
-        free(r);
-    }
+    r->con = (connection *)reqpool; /*(reuse r->con as next ptr)*/
+    reqpool = r;
 }
 
 
@@ -263,7 +274,6 @@ request_acquire (connection * const con)
     request_st *r = reqpool;
     if (r) {
         reqpool = (request_st *)r->con; /*(reuse r->con as next ptr)*/
-        ++reqspace;
     }
     else {
         r = calloc(1, sizeof(request_st));
