@@ -9,8 +9,9 @@
 
 static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in, size_t in_len, const char *out, size_t out_len) {
 	buffer_copy_string_len(psrc, in, in_len);
+	pdest = psrc; /*(buffer_path_simplify() now takes only one arg)*/
 
-	buffer_path_simplify(pdest, psrc);
+	buffer_path_simplify(pdest);
 
 	if (!buffer_eq_slen(pdest, out, out_len)) {
 		fprintf(stderr,
@@ -19,12 +20,11 @@ static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in
 			__LINE__,
 			in,
 			out,
-			pdest->ptr ? pdest->ptr : "");
+			pdest->ptr);
 		fflush(stderr);
 		abort();
 	} else {
-		if (psrc != pdest) buffer_copy_buffer(psrc, pdest);
-		buffer_path_simplify(pdest, psrc);
+		buffer_path_simplify(pdest);
 
 		if (!buffer_eq_slen(pdest, out, out_len)) {
 			fprintf(stderr,
@@ -33,7 +33,7 @@ static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in
 				__LINE__,
 				in,
 				out,
-				pdest->ptr ? pdest->ptr : "");
+				pdest->ptr);
 			fflush(stderr);
 			abort();
 		}
@@ -70,14 +70,11 @@ static void test_buffer_path_simplify_with(buffer *psrc, buffer *pdest) {
 
 static void test_buffer_path_simplify(void) {
 	buffer *psrc = buffer_init();
-	buffer *pdest = buffer_init();
 
 	/* test with using the same buffer and with using different buffers */
 	test_buffer_path_simplify_with(psrc, psrc);
-	test_buffer_path_simplify_with(pdest, psrc);
 
 	buffer_free(psrc);
-	buffer_free(pdest);
 }
 
 static void test_buffer_to_lower_upper(void) {
